"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.clients.restful.config import RestConfiguration


class APIKeysHandler:
    API_KEY_HEADER = "PROGRAMY-API-KEY"
    API_KEY_ARG = 'apikey'

    def __init__(self, configuration: RestConfiguration):
        self._configuration = configuration
        self.api_keys = []

    @staticmethod
    def format_get_api_key_param(api_key):
        return "%s=%s" % (APIKeysHandler.API_KEY_ARG, api_key)

    @staticmethod
    def add_post_api_key_header(headers, api_key):
        headers['PROGRAMY-API-KEY'] = api_key

    def load_api_keys(self):
        if self._configuration.use_api_keys is True:
            if self._configuration.api_key_file is not None:
                try:
                    with open(self._configuration.api_key_file, "r", encoding="utf-8") as api_key_file:
                        for api_key in api_key_file:
                            self.api_keys.append(api_key.strip())

                except Exception as excep:
                    YLogger.exception(self, "Failed to open license key file [%s]", excep,
                                      self._configuration.api_key_file)

    def use_api_keys(self):
        return self._configuration.use_api_keys

    def get_api_key(self, rest_request, method='GET'):
        if method == 'GET':
            if APIKeysHandler.API_KEY_ARG in rest_request.args and \
                    rest_request.args[APIKeysHandler.API_KEY_ARG] is not None:
                return rest_request.args[APIKeysHandler.API_KEY_ARG]

        elif method == 'POST':
            if APIKeysHandler.API_KEY_HEADER in rest_request.headers and \
                    rest_request.headers[APIKeysHandler.API_KEY_HEADER] is not None:
                return rest_request.headers[APIKeysHandler.API_KEY_HEADER]

        return None

    def is_apikey_valid(self, apikey):
        return bool(apikey in self.api_keys)

    def verify_api_key_usage(self, request, method='GET'):
        if self.use_api_keys() is True:
            apikey = self.get_api_key(request, method)

            if apikey is None:
                YLogger.error(self, "Unauthorised access - api required but missing")
                return False

            if self.is_apikey_valid(apikey) is False:
                YLogger.error(self, "'Unauthorised access - invalid api key")
                return False

        return True



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.mappings.base import DoubleStringCharSplitCollection
from programy.storage.factory import StorageFactory


class BasePropertiesCollection(DoubleStringCharSplitCollection):

    def __init__(self):
        DoubleStringCharSplitCollection.__init__(self)

    def get_split_char(self):
        return ":"

    def has_property(self, key):
        return self.has_key(key)

    def property(self, key):
        return self.value(key)

    def add_property(self, key, value):
        if self.has_property(key):
            self.add_value(key, value)
        else:
            self.pairs.append([key, value])

    def get_storage_name(self):
        raise NotImplementedError()  # pragma: no cover

    def get_store(self, engine):
        raise NotImplementedError()  # pragma: no cover

    def _load_collection(self, storage_engine):
        store = self.get_store(storage_engine)
        store.load_all(self)

    def load(self, storage_factory):
        name = self.get_storage_name()
        if storage_factory.entity_storage_engine_available(name) is True:
            engine = storage_factory.entity_storage_engine(name)
            try:
                self._load_collection(engine)
                return True

            except Exception as e:
                YLogger.exception(self, "Failed to load %s from storage", e, name)

        return False

    def reload(self, storage_factory):
        return self.load(storage_factory)


class PropertiesCollection(BasePropertiesCollection):

    def __init__(self):
        BasePropertiesCollection.__init__(self)

    def get_storage_name(self):
        return StorageFactory.PROPERTIES

    def get_store(self, engine):
        return engine.property_store()


class DefaultVariablesCollection(BasePropertiesCollection):

    def __init__(self):
        BasePropertiesCollection.__init__(self)

    def get_storage_name(self):
        return StorageFactory.DEFAULTS

    def get_store(self, engine):
        return engine.defaults_store()

    def has_variable(self, key):
        return self.has_property(key)

    def variable(self, key):
        return self.property(key)

    def add_variable(self, key, value):
        self.add_property(key, value)


class RegexTemplatesCollection(BasePropertiesCollection):

    def __init__(self):
        BasePropertiesCollection.__init__(self)

    def get_storage_name(self):
        return StorageFactory.REGEX_TEMPLATES

    def get_store(self, engine):
        return engine.regex_store()

    def has_regex(self, key):
        return self.has_property(key)

    def regex(self, key):
        return self.property(key)

    def add_regex(self, key, value):
        self.add_property(key, value)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.storage.engine import StorageEngine
from programy.storage.stores.file.store.properties import FilePropertyStore
from programy.storage.stores.file.store.properties import FileDefaultVariablesStore
from programy.storage.stores.file.store.conversations import FileConversationStore
from programy.storage.stores.file.store.twitter import FileTwitterStore
from programy.storage.stores.file.store.sets import FileSetsStore
from programy.storage.stores.file.store.maps import FileMapsStore
from programy.storage.stores.file.store.rdfs import FileRDFStore
from programy.storage.stores.file.store.lookups import FileDenormalStore
from programy.storage.stores.file.store.lookups import FileNormalStore
from programy.storage.stores.file.store.lookups import FileGenderStore
from programy.storage.stores.file.store.lookups import FilePersonStore
from programy.storage.stores.file.store.lookups import FilePerson2Store
from programy.storage.stores.file.store.properties import FileRegexStore
from programy.storage.stores.file.store.errors import FileErrorsStore
from programy.storage.stores.file.store.duplicates import FileDuplicatesStore
from programy.storage.stores.file.store.categories import FileCategoryStore
from programy.storage.stores.file.store.learnf import FileLearnfStore
from programy.storage.stores.file.store.spelling import FileSpellingStore
from programy.storage.stores.file.store.licensekeys import FileLicenseStore
from programy.storage.stores.file.store.nodes import FilePatternNodeStore
from programy.storage.stores.file.store.nodes import FileTemplateNodeStore
from programy.storage.stores.file.store.binaries import FileBinariesStore
from programy.storage.stores.file.store.braintree import FileBraintreeStore
from programy.storage.stores.file.store.processors import FilePreProcessorsStore
from programy.storage.stores.file.store.processors import FilePostProcessorsStore
from programy.storage.stores.file.store.processors import FilePostQuestionProcessorsStore
from programy.storage.stores.file.store.usergroups import FileUserGroupStore
from programy.storage.stores.file.store.triggers import FileTriggersStore
from programy.storage.stores.file.store.oobs import FileOOBStore


class FileStorageEngine(StorageEngine):
    FILE = "file"

    def __init__(self, configuration):
        StorageEngine.__init__(self, configuration)

    def initialise(self):
        return True

    def category_store(self):
        return FileCategoryStore(self)

    def errors_store(self):
        return FileErrorsStore(self)

    def duplicates_store(self):
        return FileDuplicatesStore(self)

    def learnf_store(self):
        return FileLearnfStore(self)

    def conversation_store(self):
        return FileConversationStore(self)

    def sets_store(self):
        return FileSetsStore(self)

    def maps_store(self):
        return FileMapsStore(self)

    def rdf_store(self):
        return FileRDFStore(self)

    def denormal_store(self):
        return FileDenormalStore(self)

    def normal_store(self):
        return FileNormalStore(self)

    def gender_store(self):
        return FileGenderStore(self)

    def person_store(self):
        return FilePersonStore(self)

    def person2_store(self):
        return FilePerson2Store(self)

    def regex_store(self):
        return FileRegexStore(self)

    def property_store(self):
        return FilePropertyStore(self)

    def defaults_store(self):
        return FileDefaultVariablesStore(self)

    def twitter_store(self):
        return FileTwitterStore(self)

    def spelling_store(self):
        return FileSpellingStore(self)

    def license_store(self):
        return FileLicenseStore(self)

    def pattern_nodes_store(self):
        return FilePatternNodeStore(self)

    def template_nodes_store(self):
        return FileTemplateNodeStore(self)

    def binaries_store(self):
        return FileBinariesStore(self)

    def braintree_store(self):
        return FileBraintreeStore(self)

    def preprocessors_store(self):
        return FilePreProcessorsStore(self)

    def postprocessors_store(self):
        return FilePostProcessorsStore(self)

    def postquestionprocessors_store(self):
        return FilePostQuestionProcessorsStore(self)

    def usergroups_store(self):
        return FileUserGroupStore(self)

    def triggers_store(self):
        return FileTriggersStore(self)

    def oobs_store(self):
        return FileOOBStore(self)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.storage.entities.store import Store
from programy.storage.stores.nosql.mongo.store.mongostore import MongoStore
from programy.storage.entities.license import LicenseStore
from programy.storage.stores.nosql.mongo.dao.license import LicenseKey
from programy.utils.console.console import outputLog


class MongoLicenseKeysStore(MongoStore, LicenseStore):
    LICENSEKEYS = "licensekeys"

    def __init__(self, storage_engine):
        MongoStore.__init__(self, storage_engine)
        LicenseStore.__init__(self)

    def collection_name(self):
        return MongoLicenseKeysStore.LICENSEKEYS

    def add_licensekey(self, key_name, key):
        licensekey = LicenseKey(name=key_name, key=key)
        return self.add_document(licensekey)

    def load(self, collector, name=None):
        del name
        self.load_all(collector)

    def load_all(self, collector):
        YLogger.info(self, "Loading license keys from Mongo")
        collection = self.collection()

        collector.empty()

        licensekeys = collection.find()
        for licensekey in licensekeys:
            collector.add_key(licensekey['name'], licensekey['key'])

    def _read_lines_from_file(self, filename, verbose):
        success = 0
        count = 0
        with open(filename, "r", encoding="utf-8") as license_file:
            for line in license_file:
                if self._process_line(line, verbose) is True:
                    success += 1
                count += 1
        return count, success

    def _process_line(self, line, verbose=False):
        line = line.strip()
        result = False
        if line and line.startswith('#') is False:
            splits = line.split("=")
            if len(splits) > 1:
                key_name = splits[0].strip()
                # If key has = signs in it, then combine all elements past the first
                key = "".join(splits[1:]).strip()
                result = self.add_licensekey(key_name, key)
                if verbose is True:
                    outputLog(self, "%s = %s" % (key_name, key))        # pragma: no cover

            else:
                YLogger.warning(self, "Invalid license key [%s]", line)

        return result

    def upload_from_file(self, filename, fileformat=Store.TEXT_FORMAT, commit=True, verbose=False):
        YLogger.info(self, "Uploading license keys to Mongo [%s]", filename)
        try:
            YLogger.info(self, "Loading license key file: [%s]", filename)
            return self._read_lines_from_file(filename, verbose)

        except Exception as excep:
            YLogger.exception(self, "Invalid license key file [%s]", excep, filename)

        return 0, 0




"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import datetime
import csv
import os
from programy.utils.logging.ylogger import YLogger


class TextFile:

    def __init__(self, filename, mode="a", encoding="utf-8"):
        self._filename = filename
        self._encoding = encoding
        self._file = open(self._filename, mode, encoding=self._encoding)

    @property
    def filename(self):
        return self._filename

    @property
    def encoding(self):
        return self._encoding

    def write_line(self, file_writer, elements):
        string = file_writer.format_row_as_text(elements)
        self._file.write(string)

    def flush(self):
        self._file.flush()


class CSVFile:

    def __init__(self, filename, mode="a", encoding="utf-8", delimiter=',', quotechar='"', quoting=csv.QUOTE_ALL):
        self._filename = filename
        self._encoding = encoding
        self._file = open(self._filename, mode, encoding=self._encoding)
        self._csv_writer = csv.writer(self._file, delimiter=delimiter, quotechar=quotechar, quoting=quoting)

    @property
    def filename(self):
        return self._filename

    @property
    def encoding(self):
        return self._encoding

    def write_line(self, _, elements):
        self._csv_writer.writerow(elements)

    def flush(self):
        return  # Do nothing, noting to flush


class FileWriterConfiguration:

    def __init__(self, filename, fileformat=None, mode="a", encoding="utf-8", delete_on_start=False):
        self._filename = filename
        self._file_format = fileformat
        self._mode = mode
        self._encoding = encoding
        self._delete_on_start = delete_on_start

    @property
    def filename(self):
        return self._filename

    @property
    def file_format(self):
        return self._file_format

    @property
    def mode(self):
        return self._mode

    @property
    def encoding(self):
        return self._encoding

    @property
    def delete_on_start(self):
        return self._delete_on_start


class FileWriter:

    def __init__(self, configuration):
        self._filename = configuration.filename

        if configuration.delete_on_start:
            if os.path.exists(configuration.filename):
                YLogger.info(self, "Removing %s on start up", configuration.filename)
                os.remove(configuration.filename)

        if configuration.file_format == 'txt':
            self._file = TextFile(self._filename, encoding=configuration.encoding)

        elif configuration.file_format == 'csv':
            self._file = CSVFile(self._filename)
            self.write_header()

        else:
            raise Exception("Unknown file type [%s]" % configuration.fileformat)

    def write_header(self):
        pass    # pragma: no cover

    def format_row_as_text(self, row):
        return row

    @staticmethod
    def get_timestamp():
        return "{:%Y-%m-%d %H:%M:%S}".format(datetime.datetime.now())


class ConversationFileWriter(FileWriter):

    def __init__(self, configuration):
        FileWriter.__init__(self, configuration)
        self.write_header()

    def log_question_and_answer(self, clientid, question, answer):
        row = [FileWriter.get_timestamp(), clientid, question, answer]
        self._file.write_line(self, row)
        self._file.flush()

    def format_row_as_text(self, row):
        return "%s - %s - Question[%s], Response[%s]\n" % (row[0], row[1], row[2], row[3])

    def write_header(self):
        self._file.write_line(self, ["Timestamp", "Clientid", "Question", "Response"])
        self._file.flush()


class ContentFileWriter(FileWriter):

    def __init__(self, configuration, content_type):
        FileWriter.__init__(self, configuration)
        self._content_type = content_type
        self._entries = []
        self.write_header()

    @property
    def entries(self):
        return self._entries

    def save_entry(self, content, filename, startline=None, endline=None):
        timestamp = self.get_timestamp()
        if startline and endline:
            row = [timestamp, content, filename, startline, endline]
        else:
            row = [timestamp, content, filename]
        self._entries.append(row)

    def save_content(self):
        YLogger.info(self, "Saving aiml %s to file [%s]", self._content_type, self._filename)

        for entry in self._entries:
            self._file.write_line(self, entry)

        self._file.flush()
        return len(self._entries)

    def format_row_as_text(self, row):
        if len(row) == 5:
            return "%s [%s] [%s] [%s] [%s]\n" % (row[0], row[1], row[2], row[3], row[4])
        else:
            return "%s [%s]\n" % (row[0], row[1])

    def display_debug_info(self):
        YLogger.info(self, "Found a total of %d %s in your grammrs, check out [%s] for details",
                     len(self._entries), self._content_type, self._filename)


class ErrorsFileWriter(ContentFileWriter):

    def __init__(self, configuration):
        ContentFileWriter.__init__(self, configuration, "errors")

    def write_header(self):
        self._file.write_line(self, ["Timestamp", "Error", "File", "Start Line", "End Line"])
        self._file.flush()


class DuplicatesFileWriter(ContentFileWriter):
    def __init__(self, configuration):
        ContentFileWriter.__init__(self, configuration, "duplicates")

    def write_header(self):
        self._file.write_line(self, ["Timestamp", "Duplicate", "File", "Start Line", "End Line"])
        self._file.flush()



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import datetime
from dateutil.relativedelta import relativedelta
from programy.utils.logging.ylogger import YLogger
from programy.parser.template.nodes.base import TemplateNode
from programy.utils.text.text import TextUtils
from programy.parser.template.nodes.word import TemplateWordNode
from programy.parser.exceptions import ParserException


class TemplateIntervalNode(TemplateNode):

    def __init__(self):
        TemplateNode.__init__(self)
        self._interval_format = None
        self._style = None
        self._interval_from = None
        self._interval_to = None

    @property
    def interval_format(self) -> TemplateNode:
        return self._interval_format

    @interval_format.setter
    def interval_format(self, interval_format: TemplateNode):
        if isinstance(interval_format, TemplateNode):
           self._interval_format = interval_format
        else:
            self._interval_format = TemplateWordNode(interval_format)

    @property
    def interval_from(self) -> TemplateNode:
        return self._interval_from

    @interval_from.setter
    def interval_from(self, interval_from):
        if isinstance(interval_from, TemplateNode):
            self._interval_from = interval_from
        else:
            self._interval_from = TemplateWordNode(interval_from)

    @property
    def interval_to(self) -> TemplateNode:
        return self._interval_to

    @interval_to.setter
    def interval_to(self, interval_to):
        if isinstance(interval_to, TemplateNode):
            self._interval_to = interval_to
        else:
            self._interval_to = TemplateWordNode(interval_to)

    @property
    def style(self) -> TemplateNode:
        return self._style

    @style.setter
    def style(self, style):
        if isinstance(style, TemplateNode):
            self._style = style
        else:
            self._style = TemplateWordNode(style)

    def resolve_to_string(self, client_context):
        if self.interval_format is not None:
            format_str = self.interval_format.resolve(client_context)
        else:
            format_str = "%c"

        from_str = self.interval_from.resolve(client_context)
        from_time = datetime.datetime.strptime(from_str, format_str)

        to_str = self.interval_to.resolve(client_context)
        to_time = datetime.datetime.strptime(to_str, format_str)

        diff = to_time - from_time
        difference = relativedelta(to_time, from_time)

        resolved = ""
        style = self.style.resolve(client_context)

        if style == "years":
            resolved = str(difference.years)
        elif style == "months":
            resolved = str(difference.months)
        elif style == "weeks":
            resolved = str(difference.weeks)
        elif style == "days":
            resolved = str(diff.days)
        elif style == "hours":
            resolved = str(difference.hours)
        elif style == "minutes":
            resolved = str(difference.minutes)
        elif style == "seconds":
            resolved = str(difference.seconds)
        elif style == "microseconds":
            resolved = str(difference.microseconds)
        elif style == "ymd":
            resolved = "%d years, %d months, %d days" % \
                       (difference.years, difference.months, difference.days)
        elif style == "hms":
            resolved = "%d hours, %d minutes, %d seconds" % \
                       (difference.hours, difference.minutes, difference.seconds)
        elif style == "ymdhms":
            resolved = "%d years, %d months, %d days, %d hours, %d minutes, %d seconds" % \
                       (difference.years, difference.months, difference.days,
                        difference.hours, difference.minutes, difference.seconds)
        else:
            YLogger.error(client_context, "Unknown interval style [%s]", style)
            resolved = ""

        YLogger.debug(client_context, "[INTERVAL] resolved to [%s]", resolved)
        return resolved

    def to_string(self):
        return "[INTERVAL]"

    def to_xml(self, client_context):
        xml = '<interval'
        if self.interval_format is not None:
            xml += ' format="%s"' % self.interval_format.to_xml(client_context)
        if self.style is not None:
            xml += ' style="%s"' % self.style.to_xml(client_context)
        xml += '>'
        if self.interval_from is not None:
            xml += '<from>'
            xml += self.interval_from.to_xml(client_context)
            xml += '</from>'
        if self.interval_to is not None:
            xml += '<to>'
            xml += self.interval_to.to_xml(client_context)
            xml += '</to>'
        xml += '</interval>'
        return xml

    #######################################################################################################
    # INTERVAL_EXPRESSION ::== <interval>
    # 							(DATE_ATTRIBUTE_TAGS)
    # 							<style>(TEMPLATE_EXPRESSION)</style>
    # 							<from>(TEMPLATE_EXPRESSION)</from>
    # 							<to>(TEMPLATE_EXPRESSION)</to>
    # 						</interval>

    def parse_expression(self, graph, expression):

        if 'format' in expression.attrib:
            self._interval_format = graph.get_word_node(expression.attrib['format'])

        if 'style' in expression.attrib:
            self._style = graph.get_word_node(expression.attrib['style'])

        if 'from' in expression.attrib:
            self._interval_from = graph.get_word_node(expression.attrib['from'])

        if 'to' in expression.attrib:
            self._interval_to = graph.get_word_node(expression.attrib['to'])

        head_text = self.get_text_from_element(expression)
        self.parse_text(graph, head_text)

        for child in expression:
            tag_name = TextUtils.tag_from_text(child.tag)

            if tag_name == 'format':
                self._interval_format = graph.get_word_node(self.get_text_from_element(child))

            elif tag_name == 'style':
                node = graph.get_base_node()
                node.parse_text(graph, self.get_text_from_element(child))
                for sub_child in child:
                    graph.parse_tag_expression(sub_child, node)
                    node.parse_text(graph, self.get_text_from_element(child))
                self._style = node

            elif tag_name == 'from':
                node = graph.get_base_node()
                node.parse_text(graph, self.get_text_from_element(child))
                for sub_child in child:
                    graph.parse_tag_expression(sub_child, node)
                    node.parse_text(graph, self.get_text_from_element(child))
                self._interval_from = node

            elif tag_name == 'to':
                node = graph.get_base_node()
                node.parse_text(graph, self.get_text_from_element(child))
                for sub_child in child:
                    graph.parse_tag_expression(sub_child, node)
                    node.parse_text(graph, self.get_text_from_element(child))
                self._interval_to = node
            else:
                raise ParserException("No children allows in interval")

            tail_text = self.get_tail_from_element(child)
            self.parse_text(graph, tail_text)

        if self.interval_format is None:
            YLogger.warning(self, "Interval node, format missing, defaulting to '%c'!")
            self._interval_format = TemplateWordNode("%c")

        if self.style is None:
            YLogger.warning(self, "style node, format missing, defaulting to 'days'!")
            self._style = TemplateWordNode("days")

        if self.interval_from is None:
            raise ParserException("interval_from node, format missing !")

        if self.interval_to is None:
            raise ParserException(self, "interval_to node, format missing !")



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
from programy.utils.logging.ylogger import YLogger
from programy.extensions.base import Extension


class SchedulerExtension(Extension):

    # execute() is the interface that is called from the <extension> tag in the AIML
    def execute(self, client_context, data):
        YLogger.debug(client_context, "Scheduler - [%s]", data)

        # SCHEDULE IN|EVERY X SECONDS|MINUTES|HOURS|DAYS|WEEKS TEXT|SRAI ...........

        try:
            words = data.split()
            if words[0].upper() == 'SCHEDULE':
                action = words[1].upper()
                if action in ['IN', 'EVERY']:
                    return self.execute_schedule(client_context, words)

                elif action == 'PAUSE':
                    return self.execute_pause(client_context, words)

                elif action == 'RESUME':
                    return self.execute_resume(client_context, words)

                elif action == 'STOP':
                    return self.execute_stop(client_context, words)

                elif action == 'LIST':
                    return self.execute_list(client_context, words)

                else:
                    raise Exception("Scheduler invalid action %s" % action)
            else:
                raise Exception("Scheduler invalid command, must start with SCHEDULE")

        except Exception as excep:
            YLogger.exception(client_context, "Failed to parse Scheduler command", excep)

        return 'ERR'

    def execute_schedule(self, client_context, words):
        action = words[1].upper()
        quantity = int(words[2])
        period = words[3].upper()
        if period in ['SECONDS', 'MINUTES', 'HOURS', 'DAYS', 'WEEKS']:
            texttype = words[4]
            if texttype in ['TEXT', 'SRAI']:
                if len(words) > 5:
                    text = " ".join(words[5:])
                    self.schedule(client_context, action, quantity, period, texttype, text)
                    return 'OK'

                else:
                    raise Exception('Scheduler action missing text')

            else:
                raise Exception('Scheduler action missing')

        else:
            raise Exception("Scheduler invalid period %s" % period)

    def schedule(self, client_context, when, quantity, period, texttype, text):

        if when == 'IN':
            if period == 'SECONDS':
                client_context.client.scheduler.schedule_in_n_seconds(client_context.userid, client_context.id,
                                                                      texttype, text, quantity)
            elif period == 'MINUTES':
                client_context.client.scheduler.schedule_in_n_minutes(client_context.userid, client_context.id,
                                                                      texttype, text, quantity)
            elif period == 'HOURS':
                client_context.client.scheduler.schedule_in_n_hours(client_context.userid, client_context.id, texttype,
                                                                    text, quantity)
            elif period == 'DAYS':
                client_context.client.scheduler.schedule_in_n_days(client_context.userid, client_context.id, texttype,
                                                                   text, quantity)
            else: # elif period == 'WEEKS':
                client_context.client.scheduler.schedule_in_n_weeks(client_context.userid, client_context.id, texttype,
                                                                    text, quantity)

        else: # elif when == 'EVERY':
            if period == 'SECONDS':
                client_context.client.scheduler.schedule_every_n_seconds(client_context.userid, client_context.id,
                                                                         texttype, text, quantity)
            elif period == 'MINUTES':
                client_context.client.scheduler.schedule_every_n_minutes(client_context.userid, client_context.id,
                                                                         texttype, text, quantity)
            elif period == 'HOURS':
                client_context.client.scheduler.schedule_every_n_hours(client_context.userid, client_context.id,
                                                                       texttype, text, quantity)
            elif period == 'DAYS':
                client_context.client.scheduler.schedule_every_n_days(client_context.userid, client_context.id,
                                                                      texttype, text, quantity)
            else: # period == 'WEEKS':
                client_context.client.scheduler.schedule_every_n_weeks(client_context.userid, client_context.id,
                                                                       texttype, text, quantity)

    def get_users_jobs(self, client_context):
        user_jobs = []
        jobs = client_context.client.scheduler.list_jobs()
        for _, job in jobs.items():
            if client_context.userid == job.args[1]:
                user_jobs.append(job)
        return user_jobs

    def execute_pause(self, client_context, words):
        which = words[2].upper()
        jobs = self.get_users_jobs(client_context)
        if jobs:
            if which == 'ALL':
                for job in jobs:
                    client_context.client.scheduler.pause_job(job.id)
                return 'OK'

            else:
                jobid = int(which)
                for job in jobs:
                    if jobid == job.id:
                        client_context.client.scheduler.pause_job(job.id)
                        return 'OK'

        return 'ERR'

    def execute_resume(self, client_context, words):
        which = words[2].upper()
        jobs = self.get_users_jobs(client_context)
        if jobs:
            if which == 'ALL':
                for job in jobs:
                    client_context.client.scheduler.resume_job(job.id)
                return 'OK'

            else:
                jobid = int(which)
                for job in jobs:
                    if jobid == job.id:
                        client_context.client.scheduler.resume_job(job.id)
                        return 'OK'

        return 'ERR'

    def execute_stop(self, client_context, words):
        which = words[2].upper()
        jobs = self.get_users_jobs(client_context)
        if jobs:
            if which == 'ALL':
                for job in jobs:
                    client_context.client.scheduler.stop_job(job.id)
                return 'OK'

            else:
                jobid = int(which)
                for job in jobs:
                    if jobid == job.id:
                        client_context.client.scheduler.stop_job(job.id)
                        return 'OK'

        return 'ERR'

    def execute_list(self, client_context, words):
        jobs = self.get_users_jobs(client_context)
        if jobs:
            response = "OK <olist>"
            for job in jobs:
                response += "<item>" + str(job.id) + "</item>"
            response += "</olist>"
            return response

        return 'ERR'




"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from programy.dynamic.maps.map import DynamicMap

# Code stolen from http://code.activestate.com/recipes/81611-roman-numerals/


class MapRomanToDecimal(DynamicMap):

    NAME = "ROMANTODEC"
    NUMS = ('M', 'D', 'C', 'L', 'X', 'V', 'I')
    INTS = (1000, 500, 100, 50, 10, 5, 1)

    def __init__(self, config):
        DynamicMap.__init__(self, config)

    def map_value(self, client_context, input_value):
        if not isinstance(input_value, str):
            raise TypeError("expected string, got %s" % type(input_value))
        input_value = input_value.upper()
        places = []
        for char in input_value:
            if char not in MapRomanToDecimal.NUMS:
                raise ValueError("input_value is not a valid roman numeral: %s" % input_value)

        charnum = 0
        for char in input_value:
            value = MapRomanToDecimal.INTS[MapRomanToDecimal.NUMS.index(char)]
            # If the next place holds a larger number, this value is negative.
            try:
                nextvalue = MapRomanToDecimal.INTS[MapRomanToDecimal.NUMS.index(input_value[charnum + 1])]
                if nextvalue > value:
                    value *= -1
            except IndexError:
                # there is no next place.
                pass
            places.append(value)
            charnum += 1
        total = 0
        for num in places:
            total += num
        return str(total)


class MapDecimalToRoman(DynamicMap):

    NAME = "DECTOROMAN"
    INTS = (1000, 900, 500, 400, 100, 90, 50, 40, 10, 9, 5, 4, 1)
    NUMS = ('M', 'CM', 'D', 'CD', 'C', 'XC', 'L', 'XL', 'X', 'IX', 'V', 'IV', 'I')

    def __init__(self, config):
        DynamicMap.__init__(self, config)

    def map_value(self, client_context, input_value):
        input_value = int(input_value)
        if not 0 < input_value < 4000:
            raise ValueError("Argument must be between 1 and 3999")
        result = ""
        num = 0
        for num_str in MapDecimalToRoman.NUMS:
            # No need to check for div by 0 as the sets of ints does not contain 0
            count = input_value // MapDecimalToRoman.INTS[num]
            result += num_str * count
            input_value -= MapDecimalToRoman.INTS[num] * count
            num += 1
        return result



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import json
from programy.utils.logging.ylogger import YLogger
from programy.services.openchatbot.response import OpenChatBotMeta
from programy.services.openchatbot.response import OpenChatBotStatus
from programy.services.openchatbot.response import OpenChatBotResponse


class OpenChatBotResponseParser:

    def __init__(self):
        self.status = None
        self.response = None
        self.meta = None

    def get_response(self, payload):
        if 'response' in payload:
            return OpenChatBotResponse.parse(payload['response'])
        return None

    def get_status(self, payload):
        if 'status' in payload:
            return OpenChatBotStatus.parse(payload['status'])
        return None

    def get_meta(self, payload):
        if 'meta' in payload:
            return OpenChatBotMeta.parse(payload['meta'])
        return None

    def parse_response(self, text):
        try:
            payload = json.loads(text)
            if payload is None:
                YLogger.error(None, "No json payload available")
                return False

        except Exception as exe:
            YLogger.exception(None, "Not valid json in response", exe)
            return False

        self.response = self.get_response(payload)
        if self.response is None:
            return False

        self.status = self.get_status(payload)
        if self.status is None:
            return False

        self.meta = self.get_meta(payload)

        if self.status.code == 200:
            return True

        return False



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from sqlalchemy import and_
from programy.storage.stores.sql.store.sqlstore import SQLStore
from programy.storage.entities.link import LinkStore
from programy.storage.stores.sql.dao.link import Link
from programy.utils.logging.ylogger import YLogger


class SQLLinkStore(SQLStore, LinkStore):

    def __init__(self, storage_engine):
        SQLStore.__init__(self, storage_engine)
        LinkStore.__init__(self)

    def _get_all(self):
        return self._storage_engine.session.query(Link)

    def empty(self):
        self._get_all().delete()

    def create_link(self, primary_userid, provided_key, generated_key, expires, expired=False, retry_count=0):
        link = Link(primary_user=primary_userid, generated_key=generated_key, provided_key=provided_key,
                    expires=expires, expired=expired, retry_count=retry_count)
        self._storage_engine.session.add(link)
        return link

    def get_link(self, userid):
        try:
            link = self._storage_engine.session.query(Link).filter(Link.primary_user == userid).one()
            return link

        except Exception as e:
            YLogger.exception_nostack(self, "Failed to get link", e)

        return None

    def _delete_link(self, userid):
        result = self._storage_engine.session.query(Link).filter(Link.primary_user == userid).delete()
        return bool(result==1)

    def remove_link(self, userid):
        try:
            return self._delete_link(userid)

        except Exception as excep:
            YLogger.exception_nostack(self, "Failed to remove link", excep)

        return False

    def link_exists(self, userid, provided_key, generated_key):
        try:
            self._storage_engine.session.query(Link).filter(Link.primary_user == userid,
                                                            Link.provided_key == provided_key,
                                                            Link.generated_key == generated_key).one()
            return True

        except Exception as excep:
            YLogger.exception_nostack(self, "Failed to check link exists", excep)
            return False

    def _get_link(self, id):
        return self._storage_engine.session.query(Link).filter(and_(Link.id == id)).one()

    def update_link(self, link):
        existing = self._get_link(link.id)
        if existing is not None:
            existing.primary_user = link.primary_user
            existing.generated_key = link.generated_key
            existing.provided_key = link.provided_key
            existing.expired = link.expired
            existing.expires = link.expires
            existing.retry_count = link.retry_count

            self._storage_engine.session.commit()
            return True

        return False



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.processors.processing import PreProcessorCollection
from programy.processors.processing import PostProcessorCollection
from programy.processors.processing import PostQuestionProcessorCollection
from programy.config.brain.brain import BrainConfiguration
from programy.mappings.denormal import DenormalCollection
from programy.mappings.gender import GenderCollection
from programy.mappings.maps import MapCollection
from programy.mappings.normal import NormalCollection
from programy.mappings.person import PersonCollection
from programy.mappings.person import Person2Collection
from programy.mappings.properties import PropertiesCollection
from programy.mappings.properties import RegexTemplatesCollection
from programy.mappings.properties import DefaultVariablesCollection
from programy.mappings.sets import SetCollection
from programy.dynamic.dynamics import DynamicsCollection
from programy.rdf.collection import RDFCollection
from programy.parser.aiml_parser import AIMLParser
from programy.services.service import ServiceFactory
from programy.services.openchatbot.collection import OpenChatBotCollection
from programy.dialog.tokenizer.tokenizer import Tokenizer
from programy.parser.pattern.factory import PatternNodeFactory
from programy.parser.template.factory import TemplateNodeFactory
from programy.binaries import BinariesManager
from programy.braintree import BraintreeManager
from programy.security.manager import SecurityManager
from programy.oob.handler import OOBHandler


class Brain:

    def __init__(self, bot, configuration: BrainConfiguration):

        assert bot is not None
        assert configuration is not None

        self._questions = 0

        self._bot = bot
        self._configuration = configuration

        self._binaries = BinariesManager(configuration.binaries)
        self._braintree = BraintreeManager(configuration.braintree)
        self._tokenizer = Tokenizer.load_tokenizer(configuration.tokenizer)

        self._denormal_collection = DenormalCollection()
        self._normal_collection = NormalCollection()
        self._gender_collection = GenderCollection()
        self._person_collection = PersonCollection()
        self._person2_collection = Person2Collection()
        self._rdf_collection = RDFCollection()
        self._sets_collection = SetCollection()
        self._maps_collection = MapCollection()

        self._properties_collection = PropertiesCollection()
        self._default_variables_collection = DefaultVariablesCollection()

        self._preprocessors = PreProcessorCollection()
        self._postprocessors = PostProcessorCollection()
        self._postquestionprocessors = PostQuestionProcessorCollection()
        self._pattern_factory = None
        self._template_factory = None

        self._security = SecurityManager(configuration.security)

        self._oobhandler = OOBHandler()

        self._openchatbots = OpenChatBotCollection()

        self._regex_templates = RegexTemplatesCollection()

        self._dynamics_collection = DynamicsCollection()

        self._aiml_parser = self.load_aiml_parser()

        self.load(self.configuration)

    @staticmethod
    def ylogger_type():
        return "brain"

    @property
    def num_questions(self):
        return self._questions

    @property
    def id(self):
        return self._configuration.section_name

    @property
    def bot(self):
        return self._bot

    @property
    def configuration(self):
        return self._configuration

    @property
    def aiml_parser(self):
        return self._aiml_parser

    @property
    def denormals(self):
        return self._denormal_collection

    @property
    def normals(self):
        return self._normal_collection

    @property
    def genders(self):
        return self._gender_collection

    @property
    def persons(self):
        return self._person_collection

    @property
    def person2s(self):
        return self._person2_collection

    @property
    def rdf(self):
        return self._rdf_collection

    @property
    def sets(self):
        return self._sets_collection

    @property
    def maps(self):
        return self._maps_collection

    @property
    def properties(self):
        return self._properties_collection

    @property
    def default_variables(self):
        return self._default_variables_collection

    @property
    def preprocessors(self):
        return self._preprocessors

    @property
    def postprocessors(self):
        return self._postprocessors

    @property
    def postquestionprocessors(self):
        return self._postquestionprocessors

    @property
    def pattern_factory(self):
        return self._pattern_factory

    @property
    def template_factory(self):
        return self._template_factory

    @property
    def regex_templates(self):
        return self._regex_templates

    @property
    def dynamics(self):
        return self._dynamics_collection

    @property
    def tokenizer(self):
        return self._tokenizer

    @property
    def openchatbots(self):
        return self._openchatbots

    @property
    def security(self):
        return self._security

    def load_aiml_parser(self):
        self._load_pattern_nodes()
        self._load_template_nodes()
        return AIMLParser(self)

    def load_aiml(self):
        YLogger.info(self, "Loading aiml source brain")
        self._aiml_parser.load_aiml()

    def reload_aimls(self):
        YLogger.info(self, "Loading aiml source brain")
        self._aiml_parser.empty()
        self._aiml_parser.load_aiml()

    def load(self, configuration: BrainConfiguration):

        load_aiml = True
        if self.configuration.binaries.load_binary is True:
            load_aiml = self._binaries.load_binary(self.bot.client.storage_factory)

        if load_aiml is True:
            self.load_aiml()

        if configuration.binaries.save_binary is True:
            self._binaries.save_binary(self.bot.client.storage_factory, self._aiml_parser)

        YLogger.info(self, "Loading collections")
        self.load_collections()

        YLogger.info(self, "Loading services")
        self.load_services(configuration)

        YLogger.info(self, "Loading openchat bots")
        self.load_openchatbots(configuration)

        YLogger.info(self, "Loading security services")
        self.load_security_services()

        YLogger.info(self, "Loading oob processors")
        self._oobhandler.load_oob_processors(self.bot.client.storage_factory)

        YLogger.info(self, "Loading regex templates")
        self.load_regex_templates()

        YLogger.info(self, "Loading dynamics sets, maps and vars")
        self.load_dynamics()

    def dump_brain_tree(self, client_context):
        self._braintree.dump_brain_tree(client_context)

    def load_denormals(self):
        self._denormal_collection.empty()
        self._denormal_collection.load(self.bot.client.storage_factory)

    def load_normals(self):
        self._normal_collection.empty()
        self._normal_collection.load(self.bot.client.storage_factory)

    def load_genders(self):
        self._gender_collection.empty()
        self._gender_collection.load(self.bot.client.storage_factory)

    def load_persons(self):
        self._person_collection.empty()
        self._person_collection.load(self.bot.client.storage_factory)

    def load_person2s(self):
        self._person2_collection.empty()
        self._person2_collection.load(self.bot.client.storage_factory)

    def load_properties(self):
        self._properties_collection.empty()
        self._properties_collection.load(self.bot.client.storage_factory)

    def load_default_variables(self):
        self._default_variables_collection.empty()
        self._default_variables_collection.load(self.bot.client.storage_factory)

        self._set_system_defined()

    def _set_system_defined(self):
        self.set_sentiment_scores(0.0, 0.5)

    def set_sentiment_scores(self, positivity, subjectivity):
        if self._default_variables_collection.has_variable("positivity") is False:
            self._default_variables_collection.set_value("positivity", str(positivity))

        if self._default_variables_collection.has_variable("subjectivity") is False:
            self._default_variables_collection.set_value("subjectivity", str(subjectivity))

    def load_maps(self):
        self._maps_collection.empty()
        self._maps_collection.load(self.bot.client.storage_factory)

    def reload_map(self, mapname):
        if self._maps_collection.contains(mapname):
            self._maps_collection.reload(self.bot.client.storage_factory, mapname)
        else:
            YLogger.error(self, "Unknown map name [%s], unable to reload ", mapname)

    def load_sets(self):
        self._sets_collection.empty()
        self._sets_collection.load(self.bot.client.storage_factory)

    def reload_set(self, setname):
        if self._sets_collection.contains(setname):
            self._sets_collection.reload(self.bot.client.storage_factory, setname)
        else:
            YLogger.error(self, "Unknown set name [%s], unable to reload ", setname)

    def load_rdfs(self):
        self._rdf_collection.empty()
        self._rdf_collection.load(self.bot.client.storage_factory)

    def reload_rdf(self, rdfname):
        if self._rdf_collection.contains(rdfname):
            self._rdf_collection.reload(self.bot.client.storage_factory, rdfname)
        else:
            YLogger.error(self, "Unknown rdf name [%s], unable to reload ", rdfname)

    def _load_preprocessors(self):
        self._preprocessors.empty()
        self._preprocessors.load(self.bot.client.storage_factory)

    def _load_postprocessors(self):
        self._postprocessors.empty()
        self._postprocessors.load(self.bot.client.storage_factory)

    def _load_postquestionprocessors(self):
        self._postquestionprocessors.empty()
        self._postquestionprocessors.load(self.bot.client.storage_factory)

    def _load_pattern_nodes(self):
        self._pattern_factory = PatternNodeFactory()
        self._pattern_factory.load(self.bot.client.storage_factory)

    def _load_template_nodes(self):
        self._template_factory = TemplateNodeFactory()
        self._template_factory.load(self.bot.client.storage_factory)

    def load_collections(self):
        self.load_denormals()
        self.load_normals()
        self.load_genders()
        self.load_persons()
        self.load_person2s()
        self.load_properties()
        self.load_default_variables()
        self.load_rdfs()
        self.load_sets()
        self.load_maps()
        self._load_preprocessors()
        self._load_postprocessors()
        self._load_postquestionprocessors()

    def load_services(self, configuration):
        ServiceFactory.preload_services(configuration.services)

    def load_openchatbots(self, configuration):
        self._openchatbots.load_from_configuration(configuration.openchatbots)

    def load_security_services(self):
        self._security.load_security_services(self.bot.client)

    def load_dynamics(self):
        self._dynamics_collection.load_from_configuration(self.configuration.dynamics)

    def load_regex_templates(self):
        self._regex_templates.load(self.bot.client.storage_factory)

    def pre_process_question(self, client_context, question):
        return self.preprocessors.process(client_context, question)

    def post_process_response(self, client_context, response: str):
        return self.postprocessors.process(client_context, response)

    def post_process_question(self, client_context, question: str):
        response = self.postquestionprocessors.process(client_context, question)
        if response == question:
            return None
        return response

    def authenticate_user(self, client_context):
        return self._security.authenticate_user(client_context)

    def resolve_matched_template(self, client_context, match_context):

        assert client_context is not None
        assert match_context is not None

        template_node = match_context.template_node

        YLogger.debug(client_context, "AIML Parser evaluating template [%s]", template_node.to_string())

        response = template_node.template.resolve(client_context)

        if self._oobhandler.oob_in_response(response) is True:
            response = self._oobhandler.handle(client_context, response)

        match_context.response = response

        return response

    def ask_question(self, client_context, sentence, srai=False):

        assert client_context is not None
        assert client_context.bot is not None
        assert self._aiml_parser is not None

        client_context.brain = self

        authenticated = self.authenticate_user(client_context)
        if authenticated is not None:
            return authenticated

        conversation = client_context.bot.get_conversation(client_context)

        self._questions += 1

        topic_pattern = conversation.get_topic_pattern(client_context)

        that_pattern = conversation.get_that_pattern(client_context, srai)

        match_context = self._aiml_parser.match_sentence(client_context,
                                                         sentence,
                                                         topic_pattern=topic_pattern,
                                                         that_pattern=that_pattern)

        if match_context is not None:
            match_context.sentence = sentence.text(client_context)
            return self.resolve_matched_template(client_context, match_context)

        return None



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import os
import os.path
import shutil
import json
from programy.utils.logging.ylogger import YLogger
from programy.storage.stores.file.store.filestore import FileStore
from programy.storage.entities.conversation import ConversationStore


class FileConversationStore(FileStore, ConversationStore):

    def __init__(self, storage_engine):
        FileStore.__init__(self, storage_engine)
        ConversationStore.__init__(self)

    def _get_storage_path(self):
        return self.storage_engine.configuration.conversation_storage.file

    def get_storage(self):
        return self.storage_engine.configuration.conversation_storage

    def empty(self):
        if os.path.exists(self._storage_engine.configuration.conversation_storage.dirs[0]) is True:
            shutil.rmtree(self._storage_engine.configuration.conversation_storage.dirs[0])

    def _conversations_filename(self, storage_dir, clientid, userid, ext="conv"):
        return "%s%s%s_%s.%s" % (storage_dir, os.sep, clientid, userid, ext)

    def _write_file(self, conversation_filepath, json_text):
        with open(conversation_filepath, "w+") as convo_file:
            convo_file.write(json_text)

    def store_conversation(self, client_context, conversation, commit=True):
        self._ensure_dir_exists(self._storage_engine.configuration.conversation_storage.dirs[0])

        conversation_filepath = self._conversations_filename(
            self._storage_engine.configuration.conversation_storage.dirs[0], client_context.client.id,
            client_context.userid)

        YLogger.debug(self, "Writing conversation to [%s]", conversation_filepath)

        convo_json = conversation.to_json()
        json_text = json.dumps(convo_json, indent=4)

        try:
            self._write_file(conversation_filepath, json_text)

        except Exception as excep:
            YLogger.exception_nostack(self, "Failed to write conversation file [%s]", excep, conversation_filepath)

    def _read_file(self, conversation_filepath, conversation):
        with open(conversation_filepath, "r+") as convo_file:
            json_text = convo_file.read()
            json_data = json.loads(json_text)
            conversation.create_from_json(json_data)

    def load_conversation(self, client_context, conversation):
        conversation_filepath = self._conversations_filename(
            self._storage_engine.configuration.conversation_storage.dirs[0], client_context.client.id,
            client_context.userid)

        if self._file_exists(conversation_filepath):
            try:
                self._read_file(conversation_filepath, conversation)
                return True

            except Exception as excep:
                YLogger.exception_nostack(self, "Failed to read conversation file [%s]", excep, conversation_filepath)

        return False


"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import os
import sys
import zipfile
import shutil
import wget
import nltk
import textblob.download_corpora as download_corpora
from programy.utils.console.console import outputLog

bots = {
    'alice2-y': "https://github.com/keiffster/alice2-y/archive/master.zip",
    'professor-y': "https://github.com/keiffster/professor-y/archive/master.zip",
    'rosie-y': "https://github.com/keiffster/rosie-y/archive/master.zip",
    'talk-y': "https://github.com/keiffster/talk-y/archive/master.zip",
    'y-bot': "https://github.com/keiffster/y-bot/archive/master.zip",
    'servusai-y': "https://github.com/keiffster/servusai-y/archive/master.zip",
    'template-y': "https://github.com/keiffster/template-y/archive/master.zip",
    'traintimes-y': "https://github.com/keiffster/traintimes-y/archive/master.zip"
}


class AdminTool:

    @staticmethod
    def recursive_copy(src, dest):
        """
        Copy each file from src dir to dest dir, including sub-directories.
        """
        for item in os.listdir(src):
            file_path = os.path.join(src, item)

            # if item is a file, copy it
            if os.path.isfile(file_path):
                shutil.copy(file_path, dest)

            # else if item is a folder, recurse
            else: # os.path.isdir(file_path):
                new_dest = os.path.join(dest, item)
                os.mkdir(new_dest)
                AdminTool.recursive_copy(file_path, new_dest)

    @staticmethod
    def delete_folder_contents(folder):
        shutil.rmtree(folder)

    @staticmethod
    def make_all_executable(path):
        for item in os.listdir(path):
            file_path = os.path.join(path, item)
            AdminTool.make_executable(file_path)

    @staticmethod
    def make_executable(path):
        mode = os.stat(path).st_mode
        mode |= (mode & 0o444) >> 2  # copy R bits to X
        os.chmod(path, mode)

    def list_bots(self):
        self.display("Available bots are:\n")
        for bot in bots:
            self.display("\t%s" % bot)
        if os.name == 'posix':
            self.display("\n\tTo download use 'python3 -m programy.admin.tool download <bot-name>'")
        else:                                                                                           # pragma: no cover
            self.display("\n\tTo download use 'python -m programy.admin.tool download <bot-name>'")     # pragma: no cover

        self.display("\nAdditional components are:\n")
        self.display("\ttextblob")
        if os.name == 'posix':
            self.display("\n\tTo install use 'python3 -m programy.admin.tool install <component>'")
        else:                                                                                           # pragma: no cover
            self.display("\n\tTo install use 'python -m programy.admin.tool install <component>'")      # pragma: no cover
        self.display("")

    def wget_download(self, url):
        return wget.download(url, bar=wget.bar_thermometer)     # pragma: no cover

    def download_bot(self, bot_name):
        url = bots[bot_name]
        self.display("Downloading [%s] from [%s]" % (bot_name, url))
        filename = self.wget_download(url)
        self.display("\nDownload complete")
        return filename

    @staticmethod
    def extract_bot(filename, path=None, remove_after=True):
        zip_ref = zipfile.ZipFile(filename, 'r')
        zip_ref.extractall(path=path)
        zip_ref.close()
        if remove_after is True:
            os.remove(filename)

    @staticmethod
    def zip_dir_name_from_filename(filename):
        return filename.split(".")[0]

    def download_and_make_active(self, bot_name):
        filename = self.download_bot(bot_name)                                              # pragma: no cover
        AdminTool.extract_bot(filename)                                                     # pragma: no cover
        master_dir = AdminTool.zip_dir_name_from_filename(filename)                         # pragma: no cover
        AdminTool.recursive_copy(master_dir, ".")                                           # pragma: no cover
        if os.name == 'posix':                                                              # pragma: no cover
            AdminTool.make_all_executable("." + os.sep + "scripts" + os.sep + "xnix")       # pragma: no cover
        shutil.rmtree(master_dir)                                                           # pragma: no cover

    def install_bot(self, words):

        if len(words) < 2:
            raise Exception("Missing bot name from download command")
        bot_name = words[1]

        if bot_name not in bots:
            raise Exception("Unknown bot [%s]" % bot_name)

        self.download_and_make_active(bot_name)

        self.show_execute_help(bot_name)

    def install_textblob(self):
        nltk.download('punkt')              # pragma: no cover
        nltk.download('stopwords')          # pragma: no cover
        download_corpora.download_all()     # pragma: no cover

    def install_additional(self, words):

        if len(words) < 2:
            raise Exception("Missing additional element from install command")
        additional = words[1]

        if additional == 'textblob':
            self.display("Installing additional components for %s" % additional)
            self.install_textblob()
        else:
            raise Exception("Missing additional component from install command ['textblob']")

    def show_execute_help(self, bot_name):
        self.display("\nTo run %s bot in console mode, use the following commands" % bot_name)
        if os.name == 'posix':
            self.display("\n\tcd scripts/xnix")
            self.display("\t./%s.sh" % bot_name)
        else:                                           # pragma: no cover
            self.display("\n\tcd scripts\\windows")     # pragma: no cover
            self.display("\t%s.cmd" % bot_name)         # pragma: no cover

    def show_help(self):
        self.display("Available commands are:\n")
        self.display("\thelp")
        self.display("\tlist")
        self.display("\tdownload <bot-name>")
        self.display("\tinstall <component>")
        self.display("")

    def run(self, words):
        try:
            num_words = len(words)
            if num_words > 0:
                primary = words[0]
                if primary == 'list':
                    self.list_bots()
                elif primary == 'download':
                    self.install_bot(words)
                elif primary == 'install':
                    self.install_additional(words)
                elif primary == 'help':
                    self.show_help()
                else:
                    raise Exception("Unknown primary command [%s]" % words[0])
            else:
                self.show_help()

        except Exception as excep:
            self.display(str(excep))
            self.show_help()

    def display(self, text):
        outputLog(self, text)   # pragma: no cover


if __name__ == '__main__':      # pragma: no cover
    tool = AdminTool()          # pragma: no cover
    tool.run(sys.argv[1:])      # pragma: no cover



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from flask import Flask, request, Response
from viberbot import Api
from viberbot.api.bot_configuration import BotConfiguration
from viberbot.api.viber_requests import ViberConversationStartedRequest
from viberbot.api.viber_requests import ViberFailedRequest
from viberbot.api.viber_requests import ViberMessageRequest
from viberbot.api.viber_requests import ViberSubscribedRequest
from viberbot.api.viber_requests import ViberUnsubscribedRequest
from viberbot.api.messages.text_message import TextMessage
from programy.utils.logging.ylogger import YLogger
from programy.clients.restful.flask.client import FlaskRestBotClient
from programy.clients.restful.flask.viber.config import ViberConfiguration
from programy.utils.console.console import outputLog

import time
import logging
import sched
import threading

VIBER_CLIENT = None


class ViberBotClient(FlaskRestBotClient):

    _running = False

    def __init__(self, argument_parser=None):
        self._viber_bot = None
        self._viber_token = None
        FlaskRestBotClient.__init__(self, "viber", argument_parser)

        YLogger.debug(self, "Viber Client is running....")

        self.create_bot()

    def create_bot(self):
        if self._viber_token is not None:
            self._viber_bot = self.create_viber_bot(self._viber_token)

        else:
            YLogger.error(self, "Viber token missing, unable to create client")

    def get_client_configuration(self):
        return ViberConfiguration()

    def get_license_keys(self):
        self._viber_token = self.license_keys.get_key("VIBER_TOKEN")

    def create_viber_api(self, configuration):
        return Api(configuration)

    def create_viber_bot(self, viber_token):

        if viber_token is None:
            YLogger.error(self, "'viber_token' missing")
            return None

        name = self.configuration.client_configuration.name
        if name is None:
            YLogger.error(self, "'name' missing from Viber configuration")
            return None

        avatar = self.configuration.client_configuration.avatar
        if avatar is None:
            YLogger.error(self, "'avatar' missing from Viber configuration")
            return None

        webhook = self.configuration.client_configuration.webhook
        if webhook is None:
            YLogger.error(self, "'webhook' missing from Viber configuration")
            return None

        configuration = BotConfiguration(
            name=name,
            avatar=avatar,
            auth_token=viber_token
        )

        bot = self.create_viber_api(configuration)
        if bot is not None:
            YLogger.error(self, "'Failed to create Viber api")

        return bot

    def handle_message_request(self, viber_request):
        message = viber_request.message
        userid = viber_request.sender.id

        client_context = self.create_client_context(userid)

        response = self.ask_question(userid, message)
        rendered = self.renderer.render(client_context, response)

        if self._viber_bot is not None:
            self._viber_bot.send_messages(viber_request.sender.id, [
                TextMessage(text=rendered)
            ])

    def handle_subscribed_request(self, viber_request):
        if self._viber_bot is not None:
            self._viber_bot.send_messages(viber_request.user.id, [
                TextMessage(text="Thanks for subscribing!")
            ])

    def handle_unsubscribed_request(self, viber_request):
        pass    # pragma: no cover

    def handle_conversation_started_request(self, viber_request):
        pass    # pragma: no cover

    def handle_failed_request(self, viber_request):
        pass    # pragma: no cover

    def handle_unknown_request(self, viber_request):
        YLogger.error(self, "Client failed receiving message. failure: {0}".format(viber_request))

    def receive_message(self, request):

        if self.configuration.client_configuration.debug is True:
            self.dump_request(request)

        if self._viber_bot is None:
            return Response(status=500)

        # every viber message is signed, you can verify the signature using this method
        if not self._viber_bot.verify_signature(request.get_data(), request.headers.get('X-Viber-Content-Signature')):
            return Response(status=403)

        # this library supplies a simple way to receive a request object
        viber_request = self._viber_bot.parse_request(request.get_data())

        if isinstance(viber_request, ViberMessageRequest):
            self.handle_message_request(viber_request)

        elif isinstance(viber_request, ViberSubscribedRequest):
            self.handle_subscribed_request(viber_request)

        elif isinstance(viber_request, ViberUnsubscribedRequest):
            self.handle_unsubscribed_request(viber_request)

        elif isinstance(viber_request, ViberConversationStartedRequest):
            self.handle_conversation_started_request(viber_request)

        elif isinstance(viber_request, ViberFailedRequest):
            self.handle_failed_request(viber_request)

        else:
            self.handle_unknown_request(viber_request)

        return Response(status=200)


def set_webhook(viber):
    print("Setting webhook", VIBER_CLIENT.configuration.client_configuration.webhook)
    viber._viber_bot.set_webhook(VIBER_CLIENT.configuration.client_configuration.webhook)


def scheduler_set_webhook(seconds, priority, viber):
    scheduler = sched.scheduler(time.time, time.sleep)
    scheduler.enter(seconds, priority, set_webhook, (viber,))
    t = threading.Thread(target=scheduler.run)
    t.start()


if __name__ == "__main__":

    outputLog(None, "Initiating Viber Client...")

    APP = Flask(__name__)

    VIBER_CLIENT = ViberBotClient()

    @APP.route(VIBER_CLIENT.configuration.client_configuration.api, methods=['POST'])
    def receive_message():
        try:
            print("viber calling")
            if VIBER_CLIENT is not None:
                return VIBER_CLIENT.receive_message(request)

        except Exception as e:
            YLogger.exception(None, "Viber error", e)

    scheduler_set_webhook(5, 1, VIBER_CLIENT)

    VIBER_CLIENT.run(APP)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.mappings.base import BaseCollection
from programy.storage.factory import StorageFactory


class RDFEntity:

    def __init__(self, subject):
        self._subject = subject
        self._predicates = {}

    @property
    def predicates(self):
        return self._predicates


class RDFCollection(BaseCollection):

    RDFS = "rdfs"

    def __init__(self):
        BaseCollection.__init__(self)
        self._entities = {}
        self._stores = {}
        self._entities_to_ids = {}
        self._entities_to_stores = {}

    @property
    def entities(self):
        return self._entities

    @property
    def stores(self):
        return self._stores

    @property
    def entities_to_ids(self):
        return self._entities_to_ids

    @property
    def entities_to_stores(self):
        return self._entities_to_stores

    def empty(self, rdf_name=None):
        if rdf_name is None:
            self._remove_all()
        else:
           self._remove_entities_by_rdf_name(rdf_name)

    def _remove_all(self):
        YLogger.debug(self, "Emptying RDF Collection")
        self._entities.clear()
        self._stores.clear()
        self._entities_to_ids.clear()
        self._entities_to_stores.clear()

    def _remove_entities_by_rdf_name(self, rdf_name):
        YLogger.debug(self, "Removing entities for [%s]", rdf_name)
        if rdf_name in self._entities_to_stores.values():
            for subject, name in self._entities_to_stores.items():
                if name == rdf_name:
                    self.delete_entity(subject)

        if rdf_name in self._stores:
            del self._stores[rdf_name]

    def contains(self, rdfname):
        return bool(rdfname.upper() in self._stores)

    def storename(self, mapname):
        if mapname in self._stores:
            return self._stores[mapname]
        return None

    def subjects(self):
        return list(self._entities.keys())

    def predicates(self, subject):
        if subject in self._entities:
            return list(self._entities[subject].predicates.keys())
        return []

    def objects(self, subject, predicate):
        if subject in self._entities:
            if predicate in self._entities[subject].predicates:
                return [self._entities[subject].predicates[predicate]]
        return []

    def load(self, storage_factory):
        YLogger.debug(self, "Loading RDF Collection")
        if storage_factory.entity_storage_engine_available(StorageFactory.RDF) is True:
            rdf_engine = storage_factory.entity_storage_engine(StorageFactory.RDF)
            try:
                rdfs_store = rdf_engine.rdf_store()
                rdfs_store.load_all(self)
                return True

            except Exception as e:
                YLogger.exception(self, "Failed to load rdf from storage", e)

        return False

    def reload(self, storage_factory, rdf_name):
        YLogger.debug(self, "Reloading RDF [%s]", rdf_name)
        if storage_factory.entity_storage_engine_available(StorageFactory.RDF) is True:
            rdf_engine = storage_factory.entity_storage_engine(StorageFactory.RDF)
            try:
                rdfs_store = rdf_engine.rdf_store()
                rdfs_store.reload(self, rdf_name)
                return True

            except Exception as e:
                YLogger.exception(self, "Failed to load rdf from storage", e)

        return False

    def add_entity(self, subject, predicate, obj, rdf_name, rdf_store=None, entityid=None):
        YLogger.debug(self, "Adding RDF Entity [%s] [%s] [%s] [%s]", subject, predicate, obj, rdf_name)

        subject = subject.upper()
        predicate = predicate.upper()

        if subject not in self._entities:
            the_subject = RDFEntity(subject)
            self._entities[subject] = the_subject

            if entityid is not None:
                entityid = entityid.upper()
                if entityid not in self._entities_to_ids:
                    self._entities_to_ids[entityid] = []
                self._entities_to_ids[entityid].append(the_subject)

            self._stores[rdf_name] = rdf_store
            self._entities_to_stores[subject] = rdf_name

        entity = self._entities[subject]

        if predicate in entity.predicates:
            entity.predicates[predicate].append(obj)
        else:
            entity.predicates[predicate] = [obj]

    def has_subject(self, subject):
        return bool(subject.upper() in self._entities)

    def has_predicate(self, subject, predicate):
        if self.has_subject(subject):

            entity = self._entities[subject.upper()]
            return bool(predicate.upper() in entity.predicates)

        return False

    def has_object(self, subject, predicate, obj):
        if self.has_subject(subject):
            entity = self._entities[subject.upper()]

            if self.has_predicate(subject, predicate):
                objects = entity.predicates[predicate.upper()]

                for anobject in objects:
                    if obj == anobject:
                        return True

        return False

    def delete_entity(self, subject, predicate=None, obj=None):
        YLogger.debug(self, "Deleting RDF Entity [%s] [%s] [%s]", subject, predicate, obj)

        if self.has_subject(subject):

            if predicate is None and obj is None:
                YLogger.debug(None, "Removing entire subject %s", subject)
                del self._entities[subject.upper()]
                return

            entity = self._entities[subject.upper()]

            if obj is None:
                if self.has_predicate(subject, predicate):
                    YLogger.debug(None, "Removing entire predicate %s, subject %s", predicate, subject)
                    del entity.predicates[predicate.upper()]
            else:
                if predicate.upper() in entity.predicates:
                    objects = entity.predicates[predicate.upper()]

                    if obj in objects:
                        YLogger.debug(None, "Removing object %s from predicate %s, subject %s", obj, predicate, subject)
                        objects = entity.predicates[predicate.upper()]
                        objects.remove(obj)

                    if len(entity.predicates[predicate.upper()]) == 0:
                        YLogger.debug(None, "Removing empty predicate %s, subject %s", predicate, subject)
                        del entity.predicates[predicate.upper()]

            if len(entity.predicates.keys()) == 0:
                YLogger.debug(None, "Removing empty subject %s", subject)
                del self._entities[subject.upper()]

    def all_as_tuples(self):
        tuples = []
        for subject, entity in self._entities.items():
            for predicate, objects in entity.predicates.items():
                for obj in objects:
                    tuples.append([subject, predicate, obj])
        return tuples

    def matched_as_tuples(self, subject=None, predicate=None, obj=None):
        tuples = []
        for entity_subject, entity in self._entities.items():
            if subject is None or subject == entity_subject:
                for entity_predicate, entity_objects in entity.predicates.items():
                    if predicate is None or predicate == entity_predicate:
                        for entity_obj in entity_objects:
                            if obj is None or obj == entity_obj:
                                YLogger.debug(None, "Matched tuple subject %s, predicate %s, object %s",
                                              entity_subject, entity_predicate, entity_obj)
                                tuples.append([entity_subject, entity_predicate, entity_obj])
        return tuples

    def not_matched_as_tuples(self, subject=None, predicate=None, obj=None):
        matched_tuples = self.matched_as_tuples(subject, predicate, obj)
        all_tuples = self.all_as_tuples()
        not_matched = []
        for atuple in all_tuples:
            if atuple not in matched_tuples:
                not_matched.append(atuple)

        return not_matched

    def remove(self, entities, subject=None, predicate=None, obj=None):
        if predicate is None and obj is None:
            YLogger.debug(self, "Removing subject=[%s]", subject)
        elif obj is None:
            YLogger.debug(self, "Removing subject=[%s], predicate=[%s]", subject, predicate)
        else:
            YLogger.debug(self, "Removing subject=[%s], predicate=[%s], object=[%s]", subject, predicate, obj)

        removes = []
        for entity in entities:
            if subject is not None:
                if predicate is not None:

                    if obj is not None:
                        if subject == entity[0] and predicate == entity[1] and obj == entity[2]:
                            removes.append(entity)

                    else:
                        if subject == entity[0] and predicate == entity[1]:
                            removes.append(entity)

                else:

                    if obj is not None:
                        if subject == entity[0] and obj == entity[2]:
                            removes.append(entity)

                    else:
                        if subject == entity[0]:
                            removes.append(entity)

            elif predicate is not None:

                if obj is not None:
                    if predicate == entity[1] and obj == entity[2]:
                        removes.append(entity)

                else:
                    if predicate == entity[1]:
                        removes.append(entity)

            elif obj is not None:

                if obj == entity[2]:
                    removes.append(entity)

            else:
                YLogger.warning(None, "RDF remove, no subject, predicte or object specified, nothing removed!")

        return [entity for entity in entities if entity not in removes]

    @staticmethod
    def _get_subject_element(subject, entity_subject):
        subj_element = None
        if subject is not None:

            if subject.startswith("?"):
                subj_element = [subject, entity_subject]

            elif subject == entity_subject:
                subj_element = ["subj", entity_subject]

        else:
            subj_element = ["subj", entity_subject]

        return subj_element

    @staticmethod
    def _get_predicate_element(predicate, entity_pred):
        pred_element = None
        if predicate is not None:

            if predicate.startswith("?"):
                pred_element = [predicate, entity_pred]

            elif predicate == entity_pred:
                pred_element = ["pred", entity_pred]

        else:
            pred_element = ["pred", entity_pred]

        return pred_element

    @staticmethod
    def _get_object_elements(entity, obj, entity_pred):

        obj_elements = []
        if obj is not None:

            if obj.startswith("?"):

                for objct in entity.predicates[entity_pred]:
                    obj_elements.append([obj, objct])

            else:

                for objct in entity.predicates[entity_pred]:

                    if objct == obj:
                        obj_elements.append(["obj", objct])
        else:

            for objct in entity.predicates[entity_pred]:
                obj_elements.append(["obj", objct])

        return obj_elements

    def match_to_vars(self, subject=None, predicate=None, obj=None):
        if predicate is None and obj is None:
            YLogger.debug(self, "Matching subject=[%s]", subject)

        elif obj is None:
            YLogger.debug(self, "Matching subject=[%s], predicate=[%s]", subject, predicate)

        else:
            YLogger.debug(self, "Matching subject=[%s], predicate=[%s], object=[%s]", subject, predicate, obj)

        results = []
        for entity_subject, entity in self._entities.items():
            subj_element = RDFCollection._get_subject_element(subject, entity_subject)

            for entity_pred in entity.predicates:

                pred_element = RDFCollection._get_predicate_element(predicate, entity_pred)

                obj_elements = RDFCollection._get_object_elements(entity, obj, entity_pred)

                if subj_element is not None and pred_element is not None and len(obj_elements) > 0:
                    for objct in obj_elements:
                        results.append([subj_element, pred_element, objct])

        return results

    def not_match_to_vars(self, subject=None, predicate=None, obj=None):
        if predicate is None and obj is None:
            YLogger.debug(self, "Not matching subject=[%s]", subject)
        elif obj is None:
            YLogger.debug(self, "Not matching subject=[%s], predicate=[%s]", subject, predicate)
        else:
            YLogger.debug(self, "Not matching subject=[%s], predicate=[%s], object=[%s]", subject, predicate, obj)

        if subject is not None and subject.startswith("?") is True:
            all_subject = subject
        else:
            all_subject = None

        if predicate is not None and predicate.startswith("?") is True:
            all_predicate = predicate
        else:
            all_predicate = None

        if obj is not None and obj.startswith("?") is True:
            all_obj = obj
        else:
            all_obj = None

        all_vars = self.match_to_vars(all_subject, all_predicate, all_obj)

        matched = self.match_to_vars(subject, predicate, obj)

        to_remove = []
        for entity in all_vars:
            for atuple in matched:
                if entity[0][1] == atuple[0][1]:
                    to_remove.append(entity)

        return [entity for entity in all_vars if entity not in to_remove]

    def match_only_vars(self, subject=None, predicate=None, obj=None):
        if predicate is None and obj is None:
            YLogger.debug(self, "Matching only vars subject=[%s]", subject)
        elif obj is None:
            YLogger.debug(self, "Matching only vars subject=[%s], predicate=[%s]", subject, predicate)
        else:
            YLogger.debug(self, "Matching only vars subject=[%s], predicate=[%s], object=[%s]", subject, predicate, obj)

        results = self.match_to_vars(subject, predicate, obj)
        returns = []
        for atuple in results:
            aset = []
            for pair in atuple:
                if pair[0].startswith("?"):
                    aset.append(pair)
            returns.append(aset)

        return returns

    def get_unify_vars(self, variables):
        unified_vars = {}
        for var in variables:
            unified_vars[var] = None
        return unified_vars

    def dict_to_array(self, adict):
        atuple = []
        for name, value in adict.items():
            atuple.append([name, value])
        return atuple

    def unify(self, variables, sets):
        YLogger.debug(self, "Unifying Vars [%s]", variables)

        if not sets:
            return []

        unifications = []
        # For each tuple in the first set
        for atuple in sets[0]:
            # Check to see if there are vars thay need unifying
            unified_vars = self.get_unify_vars(variables)
            if unified_vars:
                # Get the value of the tuples which match the vars
                if self.unify_tuple(atuple, unified_vars) is True:
                    # If we have more sets, keep checking that the vars are still matching
                    unified = True
                    if len(sets) > 1:
                        unified = self.unify_set(1, sets, unified_vars, 0)
                    # If we have unified all variables then we have a match
                    if unified is True and list(unified_vars.values()):
                        unifications.append(self.dict_to_array(unified_vars))

        return unifications

    def unify_set(self, num_set, sets, unified_vars, depth):
        YLogger.debug(self, "Unifying Set")

        aset = sets[num_set]
        unified = False
        for atuple in aset:
            if self.unify_tuple(atuple, unified_vars) is False:
                continue
            else:
                unified = True
            if num_set < len(sets)-1:
                return self.unify_set(num_set+1, sets, unified_vars, depth+1)

        return unified

    def unify_tuple(self, tuples, variables):
        YLogger.debug(self, "Unifying Tuple")

        for name, value in tuples:
            if name in variables:
                if variables[name] is None:
                    variables[name] = value
                else:
                    if variables[name] != value:
                        return False

        return True



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from flask import Flask, request, Response
from kik import KikApi, Configuration
from kik.messages import messages_from_json, TextMessage
from programy.utils.logging.ylogger import YLogger
from programy.clients.restful.flask.client import FlaskRestBotClient
from programy.clients.restful.flask.kik.config import KikConfiguration
from programy.utils.console.console import outputLog


class KikBotClient(FlaskRestBotClient):

    def __init__(self, argument_parser=None):
        self._bot_api_key = None
        FlaskRestBotClient.__init__(self, "kik", argument_parser)

        self._kik_bot = self.create_kik_bot()

        YLogger.debug(self, "Kik Client is running....")

    def get_client_configuration(self):
        return KikConfiguration()

    def get_license_keys(self):
        self._bot_api_key = self.license_keys.get_key("KIK_BOT_API_KEY")

    def create_kik_bot(self):
        if self._bot_api_key is not None:
            kik_bot = KikApi(self.configuration.client_configuration.bot_name, self._bot_api_key)
            kik_bot.set_configuration(Configuration(webhook=self.configuration.client_configuration.webhook))
            return kik_bot

        else:
            YLogger.error(self, "Kik bot api key missing, unable to create kik bot")
            return None

    def handle_text_message(self, message):
        question = message.body
        userid = message.from_user

        client_context = self.create_client_context(userid)

        answer = self.ask_question(userid, question)
        rendered = self.renderer.render(client_context, answer)

        self._kik_bot.send_messages([
            TextMessage(
                to=message.from_user,
                chat_id=message.chat_id,
                body=rendered
            )
        ])

    def get_unknown_response(self, userid):
        if self.configuration.client_configuration.unknown_command_srai is None:
            unknown_response = self.configuration.client_configuration.unknown_command
        else:
            unknown_response = self.ask_question(userid, self.configuration.client_configuration.unknown_command_srai)
            if unknown_response is None or unknown_response == "":
                unknown_response = self.configuration.client_configuration.unknown_command

        client_context = self.create_client_context(userid)

        return self.renderer.render(client_context, unknown_response)

    def handle_unknown_message(self, message):
        userid = message.from_user

        unknown_response = self.get_unknown_response(userid)

        self._kik_bot.send_messages([
            TextMessage(
                to=message.from_user,
                chat_id=message.chat_id,
                body=unknown_response
            )
        ])

    def handle_message_request(self, request):

        messages = messages_from_json(request.json['messages'])

        for message in messages:
            if isinstance(message, TextMessage):
                self.handle_text_message(message)
            else:
                self.handle_unknown_message(message)

    def receive_message(self, request):

        if self.configuration.client_configuration.debug is True:
            self.dump_request(request)

        if self._kik_bot is None:
            return Response(status=500)

        if not self._kik_bot.verify_signature(request.headers.get('X-Kik-Signature'), request.get_data()):
            return Response(status=403)

        self.handle_message_request(request)
        return Response(status=200)


if __name__ == "__main__":

    outputLog(None, "Initiating Kik Client...")

    KIK_CLIENT = KikBotClient()

    APP = Flask(__name__)

    @APP.route(KIK_CLIENT.configuration.client_configuration.api, methods=['POST'])
    def receive_message():
        try:
            return KIK_CLIENT.receive_message(request)
        except Exception as e:
            YLogger.exception(None, "KIK Error", e)

    KIK_CLIENT.run(APP)



import os
import os.path
import csv
from optparse import OptionParser


class CmdArgsHandler:

    def validate_args(self):
        parser = OptionParser()
        parser.add_option("-f", "--file", dest="csvloc",
                          help="Specify the location and filename of the CSV file.")
        parser.add_option("-d", "--directory", dest="dir",
                          help="Specify the directory which contains the CSV.")
        parser.add_option("-o", "--output", dest="aimlloc",
                          help="Specify the location where the aiml file should go.")

        (opts, _) = parser.parse_args()

        mandatories = ['aimlloc']
        self.check_manditory_opts_present(parser, opts, mandatories)
        return opts

    def check_manditory_opts_present(self, parser, opts, mandatories):

        for option in mandatories:
            if not opts.__dict__[option]:
                print("ERROR: Mandatory argument is missing.\n")
                parser.print_help()
                exit(-1)

        if not opts.csvloc and not opts.dir:
            print("ERROR: Enter the csv file location, or the directory.\n")
            parser.print_help()
            exit(-1)


class CsvReader:

    def __init__(self, file_name):
        self.file_name = file_name
        self.csvlength = 0

    def read_file(self):
        if not os.path.exists(self.file_name):
            print(self.file_name + " is not a valid file.")
            exit(-1)

        collection = []
        with open(self.file_name, 'r') as f:
            reader = csv.reader(f)
            for linenum, row in enumerate(reader):
                if str(row).find('#') != -1:
                    continue
                if str(row).find(',') != -1:
                    collection.append([linenum + 1] + row)
                    self.csvlength += 1
        return collection


class CollectionLoader:

    def __init__(self, csv_file_name):
        self.csv_file_name = csv_file_name

    def get_collection(self):
        read = CsvReader(self.csv_file_name)
        collection = read.read_file()
        self.csvlength = read.csvlength
        return collection

    def format_collection(self, collection):
        parser = SentenceParser(collection[1:]) # omit line number
        return parser.populate()


class XmlWriter:

    def __init__(self, output_loc, file_name):
        self.aiml_file = open(output_loc + file_name + ".aiml", "w")

    def prepare_sentence(self, sentence):
        sentence.pop(0)
        value = ' '.join(sentence)
        return value

    def write_body(self, sentence):
        template_content = sentence[0]
        pattern_text = self.prepare_sentence(sentence)
        self.aiml_file.write("\n\t<category>")
        self.aiml_file.write("\n\t\t<pattern>")
        self.aiml_file.write(pattern_text.upper())
        self.aiml_file.write("</pattern>")
        self.aiml_file.write("\n\t\t<template>")
        self.aiml_file.write(template_content)
        self.aiml_file.write("\n\t\t</template>")
        self.aiml_file.write("\n\t</category>")

    def open_file(self):
        self.aiml_file.write("<?xml version='1.0' encoding='ISO-8859-1'?>")
        self.aiml_file.write("\n<aiml version=\"1.0.1\">")
        self.aiml_file.write("""\n
\t<!--  -->
\t<!-- This AIML file has been auto generated by the Program-Y util aiml_generator. -->
\t<!--  -->
\t<!-- Y-Bot is Copyright &copy; 2017 by Keith Sterling. -->
\t<!--
\tPermission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
\tdocumentation files (the "Software"), to deal in the Software without restriction, including without limitation
\tthe rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
\tand to permit persons to whom the Software is furnished to do so, subject to the following conditions:

\tThe above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

\tTHE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
\tTHE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
\tAUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
\tTORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
\t-->
    """)

    def close_file(self):
        self.aiml_file.write("\n</aiml>\n")
        self.aiml_file.close()


class SentenceParser:

    def __init__(self, sentence):
        self.sentence = sentence

    def get_sentence_element(self, i):
        return self.sentence[i].strip()

    def get_sentence_length(self):
        return len(self.sentence)

    def parser(self, trees, branch, count):
        if self.get_sentence_length() <= count:
            return

        word = self.get_sentence_element(count)

        # Optional word
        if "(" in word and count > 0:
            alt = branch[:]
            alt.append(word.strip("()"))
            trees.append(alt)
            self.parser(trees, branch, count + 1)
            self.parser(trees, alt, count + 1)

        # One from the list
        elif word.count('|') > 0:

            split = word.split('|')
            options = len(split)

            for i in range(1, options):
                alt = branch[:]
                alt.append(split[i])
                trees.append(alt)
                self.parser(trees, alt, count + 1)

            branch.append(split[0])
            self.parser(trees, branch, count + 1)

        # Only one word allowed by csv column
        elif (count > 0) and (word.count(' ') > 0):
            print ("WARNING: Comma missing in file [%s]" % word)
            branch.append(word)
            self.parser(trees, branch, count + 1)

        # Single word
        else:
            branch.append(word)
            self.parser(trees, branch, count + 1)

    def populate(self):
        trees = []
        branch = []
        trees.append(branch)
        self.parser(trees, branch, 0)
        return trees


class Generator():
    def __init__(self):
        self.categories = {}

    def get_aiml_file_name(self, csv_file_name):
        initial_split = csv_file_name.split('/')
        i = len(initial_split)
        secondary_split = initial_split[i - 1].split('.')

        return secondary_split[0]

    def generate_one(self, opts, csv_loc):
        cl = CollectionLoader(csv_loc)

        collections = cl.get_collection()
        if not collections:
            print(csv_loc + " has no content.")
            return
        aiml_file_name = self.get_aiml_file_name(csv_loc)

        xml = XmlWriter(opts.aimlloc, aiml_file_name)
        xml.open_file()

        for collection in collections:
            parsed_collection = cl.format_collection(collection)
            for rule in parsed_collection:
                xml.write_body(rule)


        xml.close_file()

    def get_files_in_dir(self, opts):
        if not os.path.exists(opts.dir):
            print(opts.dir + " is not a valid directory.")
            exit(-1)

        file_names = []
        for file in os.listdir(opts.dir):
            if file.endswith(".csv"):
                file_names.append(str(os.path.join(opts.dir, file)))
        if not file_names:
            print(
                "FATAL: There are no CSV files at this location.\n Are you in the right directory?")
            exit(-1)

        return file_names


if __name__ == '__main__':

    cmd = CmdArgsHandler()
    opts = cmd.validate_args()
    g = Generator()

    if not opts.aimlloc.endswith("/"):
        opts.aimlloc = opts.aimlloc + "/"

    if not os.path.exists(opts.aimlloc):
        print(opts.aimlloc + " is not a valid output directory.")
        exit(-1)

    if opts.dir:
        if not opts.dir.endswith("/"):
            opts.dir = opts.dir + "/"

        file_names = g.get_files_in_dir(opts)
        for csv_file in file_names:
            g.generate_one(opts, csv_file)
    else:
        g.generate_one(opts, opts.csvloc)




"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from programy.storage.stores.sql.store.users import SQLUserStore
from programy.storage.stores.sql.store.linkedaccounts import SQLLinkedAccountStore
from programy.storage.stores.sql.store.links import SQLLinkStore
from programy.storage.stores.sql.store.properties import SQLPropertyStore
from programy.storage.stores.sql.store.properties import SQLRegexStore
from programy.storage.stores.sql.store.properties import SQLDefaultVariableStore
from programy.storage.stores.sql.store.categories import SQLCategoryStore
from programy.storage.stores.sql.store.learnf import SQLLearnfStore
from programy.storage.stores.sql.store.conversations import SQLConversationStore
from programy.storage.stores.sql.store.sets import SQLSetsStore
from programy.storage.stores.sql.store.maps import SQLMapsStore
from programy.storage.stores.sql.store.rdfs import SQLRDFsStore
from programy.storage.stores.sql.store.lookups import SQLDenormalStore
from programy.storage.stores.sql.store.lookups import SQLNormalStore
from programy.storage.stores.sql.store.lookups import SQLGenderStore
from programy.storage.stores.sql.store.lookups import SQLPersonStore
from programy.storage.stores.sql.store.lookups import SQLPerson2Store
from programy.storage.stores.sql.store.errors import SQLErrorsStore
from programy.storage.stores.sql.store.duplicates import SQLDuplicatesStore
from programy.storage.stores.sql.store.twitter import SQLTwitterStore
from programy.storage.stores.sql.store.spelling import SQLSpellingStore
from programy.storage.stores.sql.store.licensekeys import SQLLicenseKeysStore
from programy.storage.stores.sql.store.nodes import SQLPatternNodesStore
from programy.storage.stores.sql.store.nodes import SQLTemplateNodesStore
from programy.storage.stores.sql.store.processors import SQLPreProcessorsStore
from programy.storage.stores.sql.store.processors import SQLPostProcessorsStore
from programy.storage.stores.sql.store.processors import SQLPostQuestionProcessorsStore
from programy.storage.stores.sql.store.usergroups import SQLUserGroupStore
from programy.storage.stores.sql.store.triggers import SQLTriggersStore
from programy.storage.stores.sql.store.oobs import SQLOOBsStore
from programy.storage.engine import StorageEngine
from programy.storage.stores.sql.base import Base


class SQLStorageEngine(StorageEngine):

    def __init__(self, configuration):
        StorageEngine.__init__(self, configuration)
        self._session = None
        self._engine = None

    def _drop_all(self):
        Base.metadata.drop_all(self._engine)

    def _create_all(self):
        Base.metadata.create_all(self._engine)

    def _create_session(self):
        Session = sessionmaker(bind=self._engine)
        self._session = Session()

    def initialise(self):
        self._engine = create_engine(self.configuration.url,
                                     encoding=self.configuration.encoding,
                                     echo=self.configuration.echo)

        if self.configuration.drop_all_first is True:
            self._drop_all()

        if self.configuration.create_db is True:
            self._create_all()

        self._create_session()
        return True

    @property
    def session(self):
        return self._session

    def user_store(self):
        return SQLUserStore(self)

    def linked_account_store(self):
        return SQLLinkedAccountStore(self)

    def link_store(self):
        return SQLLinkStore(self)

    def category_store(self):
        return SQLCategoryStore(self)

    def errors_store(self):
        return SQLErrorsStore(self)

    def duplicates_store(self):
        return SQLDuplicatesStore(self)

    def learnf_store(self):
        return SQLLearnfStore(self)

    def conversation_store(self):
        return SQLConversationStore(self)

    def sets_store(self):
        return SQLSetsStore(self)

    def maps_store(self):
        return SQLMapsStore(self)

    def rdf_store(self):
        return SQLRDFsStore(self)

    def denormal_store(self):
        return SQLDenormalStore(self)

    def normal_store(self):
        return SQLNormalStore(self)

    def gender_store(self):
        return SQLGenderStore(self)

    def person_store(self):
        return SQLPersonStore(self)

    def person2_store(self):
        return SQLPerson2Store(self)

    def regex_store(self):
        return SQLRegexStore(self)

    def property_store(self):
        return SQLPropertyStore(self)

    def defaults_store(self):
        return SQLDefaultVariableStore(self)

    def twitter_store(self):
        return SQLTwitterStore(self)

    def spelling_store(self):
        return SQLSpellingStore(self)

    def license_store(self):
        return SQLLicenseKeysStore(self)

    def pattern_nodes_store(self):
        return SQLPatternNodesStore(self)

    def template_nodes_store(self):
        return SQLTemplateNodesStore(self)

    def preprocessors_store(self):
        return SQLPreProcessorsStore(self)

    def postprocessors_store(self):
        return SQLPostProcessorsStore(self)

    def postquestionprocessors_store(self):
        return SQLPostQuestionProcessorsStore(self)

    def usergroups_store(self):
        return SQLUserGroupStore(self)

    def triggers_store(self):
        return SQLTriggersStore(self)

    def oobs_store(self):
        return SQLOOBsStore(self)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import datetime
import uuid
from programy.utils.logging.ylogger import YLogger
from programy.utils.console.console import outputLog


class ClientContext:

    def __init__(self, client, userid):
        self._client = client
        self._userid = userid
        self._bot = None
        self._brain = None
        self._question_start_time = None
        self._question_depth = 0
        self._id = uuid.uuid1()

    @staticmethod
    def ylogger_type():
        return "context"

    @property
    def id(self):
        return self._id

    @property
    def client(self):
        return self._client

    @property
    def userid(self):
        return self._userid

    @property
    def bot(self):
        return self._bot

    @bot.setter
    def bot(self, botid):
        self._bot = botid

    @property
    def brain(self):
        return self._brain

    @brain.setter
    def brain(self, brainid):
        self._brain = brainid

    def check_max_recursion(self):
        if self.bot.configuration.max_question_recursion != -1:
            if self._question_depth > self.bot.configuration.max_question_recursion:
                raise Exception("Maximum recursion limit [%d] exceeded" % self.bot.configuration.max_question_recursion)

    def total_search_time(self):
        delta = datetime.datetime.now() - self._question_start_time
        return abs(delta.total_seconds())

    def check_max_timeout(self):
        if self.bot.configuration.max_question_timeout != -1:
            if self.total_search_time() >= self.bot.configuration.max_question_timeout:
                raise Exception("Maximum search time limit [%d] exceeded" % self.bot.configuration.max_question_timeout)

    def mark_question_start(self, question):
        YLogger.debug(self,
                      "##########################################################################################")
        YLogger.debug(self, "Question (%s): %s", self._client.id, question)

        if self._question_depth == 0:
            self._question_start_time = datetime.datetime.now()

        self._question_depth += 1

    def reset_question(self):
        self._question_depth = 0

    def __str__(self):
        return "[%s] [%s] [%s] [%s] [%d]" % (
            self._client.id,
            self._userid,
            self._bot.id if self._bot else "",
            self._brain.id if self._brain else "",
            self._question_depth
        )

    def to_json(self):
        return {
            "clientid": self._client.id,
            "userid": self._userid,
            "botid": self._bot.id if self._bot else None,
            "brainid": self._brain.id if self._brain else None,
            "depth": self._question_depth
        }

    def dump(self, output_func=outputLog, verbose=False):
        """
        Helper Function for output tree structure
        :param output_func: Function used to display output, typically print or logging function
        :param verbose: Level of detail to display
        :return: None
        """
        del verbose
        self.brain.aiml_parser.pattern_parser.dump(output_func=output_func, verbose=False)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.storage.stores.file.store.filestore import FileStore
from programy.storage.entities.lookups import LookupsStore
from programy.mappings.base import DoubleStringPatternSplitCollection


class FileLookupsStore(FileStore, LookupsStore):

    def __init__(self, storage_engine):
        FileStore.__init__(self, storage_engine)
        LookupsStore.__init__(self)

    def _get_storage_path(self):
        raise NotImplementedError()  # pragma: no cover

    def get_storage(self):
        raise NotImplementedError()  # pragma: no cover

    def _load_file_contents(self, collection, filename):
        YLogger.debug(self, "Loading lookup [%s]", filename)
        try:
            with open(filename, 'r', encoding='utf8') as my_file:
                for line in my_file:
                        splits = DoubleStringPatternSplitCollection.\
                            split_line_by_pattern(line, DoubleStringPatternSplitCollection.RE_OF_SPLIT_PATTERN)
                        if splits and len(splits) > 1:
                            index, pattern = self.process_key_value(splits[0].upper(), splits[1])
                            collection.add_to_lookup(index, pattern)

        except Exception as excep:
            YLogger.exception_nostack(self, "Failed to load lookup [%s]", excep, filename)


class FileDenormalStore(FileLookupsStore):

    def __init__(self, storage_engine):
        FileLookupsStore.__init__(self, storage_engine)

    def _get_storage_path(self):
        return self.storage_engine.configuration.denormal_storage.file

    def get_storage(self):
        return self.storage_engine.configuration.denormal_storage


class FileNormalStore(FileLookupsStore):

    def __init__(self, storage_engine):
        FileLookupsStore.__init__(self, storage_engine)

    def _get_storage_path(self):
        return self.storage_engine.configuration.normal_storage.file

    def get_storage(self):
        return self.storage_engine.configuration.normal_storage


class FileGenderStore(FileLookupsStore):

    def __init__(self, storage_engine):
        FileLookupsStore.__init__(self, storage_engine)

    def _get_storage_path(self):
        return self.storage_engine.configuration.gender_storage.file

    def get_storage(self):
        return self.storage_engine.configuration.gender_storage


class FilePersonStore(FileLookupsStore):

    def __init__(self, storage_engine):
        FileLookupsStore.__init__(self, storage_engine)

    def _get_storage_path(self):
        return self.storage_engine.configuration.person_storage.file

    def get_storage(self):
        return self.storage_engine.configuration.person_storage


class FilePerson2Store(FileLookupsStore):

    def __init__(self, storage_engine):
        FileLookupsStore.__init__(self, storage_engine)

    def _get_storage_path(self):
        return self.storage_engine.configuration.person2_storage.file

    def get_storage(self):
        return self.storage_engine.configuration.person2_storage



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.storage.stores.nosql.mongo.store.mongostore import MongoStore
from programy.storage.entities.linked import LinkedAccountStore
from programy.storage.stores.nosql.mongo.dao.linked import LinkedAccount


class MongoLinkedAccountStore(MongoStore, LinkedAccountStore):
    LINKEDACCOUNTS = 'linkedaccounts'
    PRIMARY_USERID = "primary_userid"
    LINKED_USERID = "linked_userid"

    def __init__(self, storage_engine):
        MongoStore.__init__(self, storage_engine)
        LinkedAccountStore.__init__(self)

    def collection_name(self):
        return MongoLinkedAccountStore.LINKEDACCOUNTS

    def link_accounts(self, primary_userid, linked_userid):
        YLogger.info(self, "Linking accounts [%s] [%s] in Mongo", primary_userid, linked_userid)
        linked = LinkedAccount(primary_userid, linked_userid)
        return self.add_document(linked)

    def unlink_accounts(self, primary_userid):
        YLogger.info(self, "Unlinking accounts [%s] in Mongo", primary_userid)
        collection = self.collection()
        result = collection.delete_many({MongoLinkedAccountStore.PRIMARY_USERID: primary_userid})
        return bool(result.deleted_count > 0)

    def linked_accounts(self, primary_userid):
        collection = self.collection()
        linked_accounts = collection.find({MongoLinkedAccountStore.PRIMARY_USERID: primary_userid})
        accounts = []
        for account in linked_accounts:
            accounts.append(account['linked_userid'])
        return accounts

    def primary_account(self, linked_userid):
        collection = self.collection()
        account = collection.find_one({MongoLinkedAccountStore.LINKED_USERID: linked_userid})
        if account is not None:
            return account['primary_userid']
        return None



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.storage.stores.sql.store.sqlstore import SQLStore
from programy.storage.entities.user import UserStore
from programy.storage.stores.sql.dao.user import User


class SQLUserStore(SQLStore, UserStore):

    def __init__(self, storage_engine):
        SQLStore.__init__(self, storage_engine)
        UserStore.__init__(self)

    def _get_all(self):
        return self._storage_engine.session.query(User)

    def empty(self):
        self._get_all().delete()

    def add_user(self, userid, clientid):
        user = User(userid=userid, client=clientid)
        self._storage_engine.session.add(user)
        return user

    def exists(self, userid, clientid):
        try:
            self._storage_engine.session.query(User).filter(User.userid == userid, User.client == clientid).one()
            return True
        except Exception:
            return False

    def get_links(self, userid):
        links = []
        db_users = self._storage_engine.session.query(User).filter(User.userid == userid)
        for user in db_users:
            links.append(user.client)
        return links

    def _remove_user_from_db(self, userid, clientid):
        rowcount = self._storage_engine.session.query(User).filter(User.userid == userid,
                                                                   User.client == clientid).delete()
        return bool(rowcount > 0)

    def remove_user(self, userid, clientid):
        try:
            return self._remove_user_from_db(userid, clientid)

        except Exception as excep:
            YLogger.exception_nostack(self, "Failed to remove user", excep)

        return False

    def _remove_user_from_all_clients_from_db(self, userid):
        rowcount = self._storage_engine.session.query(User).filter(User.userid == userid).delete()
        return bool(rowcount > 0)

    def remove_user_from_all_clients(self, userid):
        try:
            return self._remove_user_from_all_clients_from_db(userid)

        except Exception as excep:
            YLogger.exception_nostack(self, "Failed to remove user from all clients", excep)

        return False
